#!/usr/bin/env python

"""
Bogroll
Copyright 2009-11 David North <david-bogroll@dnorth.net>
This program is distributed under the terms of the GNU General Public License.

This file is part of Bogroll.

Bogroll is free software: you can redistribute it and/or modify
it under the terms of the GNU General Public License as published by
the Free Software Foundation, either version 3 of the License, or
(at your option) any later version.

Bogroll is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with Bogroll.  If not, see <http://www.gnu.org/licenses/>.
"""

import os
import sys
import time
import pickle
import socket
import codecs
import datetime
import feedparser

from config import *
from traceback import format_exc
from Cheetah.Template import Template

# revision number for the footer
VERSION = '0.3-dev'

cacheFile = os.path.join(CACHE_DIRECTORY, 'feed-cache')

socket.setdefaulttimeout(25) # 25 seconds

def die(msg):
    """Oh dear, it's all turned to custard"""
    print >> sys.stderr, msg, "Exiting."
    sys.exit(1)

def load_cache():
    """Load or create the etag/last-modified cache and return it"""
    if not os.path.isdir(CACHE_DIRECTORY):
        try:
            os.mkdir(CACHE_DIRECTORY)
        except:
            die("Unable to create cache directory.")

    if not os.access(CACHE_DIRECTORY, os.W_OK):
        die("Cache directory is not writeable.")

    cache = None
    if os.path.isfile(cacheFile):
        try:
            cache = pickle.load(open(cacheFile))
        except:
            pass

    if not cache:
        cache = {} # maps url to tuple of (etag or last-modified date, feed items)

    return cache


def save_cache(cache):
    """Save updated cache to disk"""
    pickle.dump(cache, open(cacheFile, 'w'))


def get_sorted_articles(feed, friendly_name, category):
    """Return a list of dicts representing salient information about the most recent
       MAX_ARTICLES_PER_CATRGORY articles in feed, sorted by date descending"""
    articles = []
    rawArticles = feed['items']
    print "Feed at %s has %s articles" % (feed.url, len(rawArticles))

    def sortKey(art):
        """Not all articles seem to have this field all the time...tedious"""
        if art.has_key('updated_parsed'):
            return time.mktime(art['updated_parsed'])
        return 999999 # assume old

    rawArticles.sort(key=sortKey, reverse=True)

    for i in rawArticles[:MAX_ARTICLES_PER_CATEGORY]: # this feed can contribute at most this many articles
        a = {}
        for attr in ['title', 'link', 'description', 'author', 'content']:
            try:
                a[attr] = i[attr]
            except:
                a[attr]= "<unknown>"

        a['datetime'] = time.strftime('%r, %A %d %B %Y', i['updated_parsed'])
        a['_datetime'] = time.mktime(i['updated_parsed'])
        a['source'] = friendly_name
        a['categories'] = category
        articles.append(a)

    return articles


def render_to_template(categories, articles):
    """Generate the page from the template"""
    update_time = datetime.datetime.now()
    update_time_str = update_time.strftime('%A, %d %B %Y at %r')
    format = "%a, %d %b %Y %H:%M:%S %Z" # Tue, 14 Mar 2000 12:45:26 GMT
    next_update_time = (update_time + datetime.timedelta(minutes=UPDATE_INTERVAL)).strftime(format)

    template_text = ''.join(open(TEMPLATE,'r').readlines())
    namespace = {'version': VERSION,
                 'update_time': update_time_str,
                 'title': TITLE,
                 'articles': articles,
                 'categories': categories,
                 'default_category': DEFAULT_CATEGORY,
                 'next_update_time': next_update_time,
                }
    t = Template(template_text, searchList=[namespace])
    result = unicode(t.respond())
    out_file = codecs.getwriter('utf-8')(open(OUTFILE,'w'))
    out_file.write(result)
    out_file.close()


def fetch_feed(url, cache):
    """Fetch feed from url, respecting and updating the cache"""
    if url in cache:
        etagOrModified, feed = cache[url]
        if isinstance(etagOrModified, str): # etag
            print "Found etag for: " + url
            d = feedparser.parse(url, etag=etagOrModified)
        else: # last-modified date/time tuple
            print "Found last-modified for: " + url
            d = feedparser.parse(url, modified=etagOrModified)
    else:
        print "No recent etag or last-modified recorded for: " + url
        d = feedparser.parse(url)

    if not hasattr(d, 'status'):
        raise Exception('Unable to fetch feed (socket timeout?)')

    if 304 == d.status:
        print "Cache hit: " + url
        d = feed

    if 500 <= d.status <= 599:
        raise Exception('HTTP status code %s fetching %s' % (d.status, url))

    # update the cache
    if hasattr(d, 'etag'):
        cache[url] = (d.etag, d)
    elif hasattr(d, 'modified'):
        cache[url] = (d.modified, d)

    return d


def main():
    """Entry point method"""
    cache = load_cache()
    categories_to_articles = {}
    for friendly_name, url, category in FEEDS:
        try:
            if not categories_to_articles.has_key(category):
                categories_to_articles[category] = []

            feed = fetch_feed(url, cache)
            articles = get_sorted_articles(feed, friendly_name, category)
            categories_to_articles[category].extend(articles)
        except Exception, e:
            print >> sys.stderr, "Problem with feed at " + url + " " + format_exc(e)

    categories = categories_to_articles.keys()
    articles = []

    # restrict to the most recent MAX_ARTICLES_PER_CATEGORY in each category
    for cat in categories_to_articles:
        arts = categories_to_articles[cat]
        print "Found %s articles in category %s" % (len(arts), cat)
        arts.sort(key=lambda a: a['_datetime'], reverse=True)
        articles.extend(arts[:MAX_ARTICLES_PER_CATEGORY])

    print "Rendering..."
    render_to_template(categories, articles)
    print "Done. Saving cache."
    save_cache(cache)
    print "Exiting normally"


if __name__ == '__main__':
    main()
